/******************************************************************************
*									      *
*	File:     dgpsip.c						      *
*	Author:   Wolfgang Rupprecht <wolfgang@capsicum.wsrcc.com>	      *
*	Created:  Sun Jan 24 20:36:39 PST 1999				      *
*	Contents: DGPS over ip -- get RTCM signals from a remote tcp source   *
*									      *
*	$Id: dgpsip.c,v 1.8 1999/01/26 10:57:09 wolfgang Exp $
******************************************************************************/

/*
 *  Copyright (c) 1999 Wolfgang Rupprecht
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  			    NO WARRANTY
 *
 *    BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO
 *  WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE
 *  LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
 *  HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT
 *  WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT
 *  NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE
 *  QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
 *  PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY
 *  SERVICING, REPAIR OR CORRECTION.
 *
 *    IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
 *  WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY
 *  MODIFY AND/OR REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE
 *  LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL,
 *  INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR
 *  INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
 *  DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU
 *  OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY
 *  OTHER PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN
 *  ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>		/* for serial hacking */
#include <fcntl.h>
#include <sys/termios.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <poll.h>
#include <signal.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/filio.h>
#include <netinet/tcp.h>

#define BUFSZ 256

char           *version = "$Id: dgpsip.c,v 1.8 1999/01/26 10:57:09 wolfgang Exp $";
int             verbose = 0;

void            usage(void);
int             openserial(char *tty, int blocksz);
void            fullwrite(int fd, unsigned char *buf, int cnt);
int             tcpopen(char *host, char *servp);

char           *nowarranty = "\n\
  BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY\n\
FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN\n\
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES\n\
PROVIDE THE PROGRAM \" AS IS \" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED\n\
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF\n\
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS\n\
TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE\n\
PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,\n\
REPAIR OR CORRECTION.\n\
\n\
  IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING\n\
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR\n\
REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,\n\
INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING\n\
OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED\n\
TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY\n\
YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER\n\
PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE\n\
POSSIBILITY OF SUCH DAMAGES.\n\n\
";

int
main(int argc,
     char **argv)
{
    char           *ttyout = "/dev/gps",
                   *dgpshost = "dgps.wsrcc.com",
                   *dgpsport = "rtcm-sc104",
                    buf[BUFSZ];
    int             cnt,
                    c,
                    infd = -1,
                    outfd = -1,
                    slp = 1,
                    wflag = 0;

    while ((c = getopt(argc, argv, "h:o:p:v:w")) != EOF) {
	switch (c) {
	case 'h':		/* dgps host */
	    dgpshost = optarg;
	    break;

	case 'p':		/* dgps port */
	    dgpsport = optarg;
	    break;

	case 'o':		/* gps serial ttyout */
	    ttyout = optarg;
	    break;

	case 'v':		/* verbose */
	    verbose = atoi(optarg);
	    break;

	case 'w':		/* turn off nowarranty printf at startup */
	    wflag = 1;
	    break;

	case '?':
	default:
	    usage();
	    break;
	}
    }

    if (verbose >= 1)
	fprintf(stderr, "Version: %s\n", version);

    if (wflag == 0) {
	fprintf(stderr, "%s", nowarranty);
    }
    outfd = openserial(ttyout, 1);
    if (outfd < 0) {
	perror("open tty");
	exit(1);
    }
    while (infd < 0) {
	infd = tcpopen(dgpshost, dgpsport);
	if (infd < 0) {
	    fprintf(stderr,
	       "Can't open network input stream to %s:%s sleeping %d sec\n",
		    dgpshost, dgpsport, slp);
	    sleep(slp);
	    slp *= 2;
	    if (slp > 60)
		slp = 60;
	} else {
	    slp = 1;
	    while ((cnt = read(infd, buf, sizeof(buf))) > 0) {
		fullwrite(outfd, buf, cnt);
		if (verbose >= 10) {
		    fprintf(stderr, "Read %d bytes\n", cnt);
		} else if (verbose >= 2) {
		    fprintf(stderr, ".");
		}
	    }
	    close(infd);
	    infd = -1;
	    fprintf(stderr,
	       "Network input stream to %s:%s Closed.  Reopening.\n",
		    dgpshost, dgpsport);
	}
    }
    exit(0);
}

int
openserial(char *tty, int blocksz)
{
    int             fd;
    struct termios  termios;

    fd = open(tty, O_RDWR | O_NONBLOCK);
    if (fd < 0) {
	perror("open");
	return (-1);
    }
    if (ioctl(fd, TIOCGETA, &termios) < 0) {
	perror("TIOGETA");
	return (-1);
    }
    termios.c_ispeed = termios.c_ospeed = 4800;
    termios.c_iflag = 0;
    termios.c_oflag = 0;	/* (ONLRET) */
    termios.c_cflag = CS8 | CLOCAL | CREAD;
    termios.c_lflag = 0;
    {
	int             cnt;

	for (cnt = 0; cnt < NCCS; cnt++)
	    termios.c_cc[cnt] = -1;
    }
    termios.c_cc[VMIN] = blocksz;
    termios.c_cc[VTIME] = 2;

    if (ioctl(fd, TIOCSETAF, &termios) < 0) {
	perror("TIOSETAF");
	return (-1);
    }
    if (fcntl(fd, F_SETFL, 0) == -1) {
	perror("fcntl: set nonblock");
    }
    return (fd);
}

void
fullwrite(int fd,
	  unsigned char *buf,
	  int cnt)
{
    int             ret;

    while (cnt > 0) {
	ret = write(fd, buf, cnt);
	if (ret > 0) {
	    cnt -= ret;
	    buf += ret;
	} else {
	    perror("error writing");
	    exit(1);
	}
    }
}

void
usage(void)
{
    fprintf(stderr, "remrtc: -v verblevel -o gpstty -h dgpshost -p dgpsport\n");
}

int
tcpopen(char *host,
	char *servp)
{
    int             sockt,
                    port;
    struct hostent *hp;
    struct servent *sp;
    struct sockaddr_in rem_sockaddr_in;

    memset((char *) &rem_sockaddr_in, 0, sizeof(rem_sockaddr_in));

    if ((rem_sockaddr_in.sin_addr.s_addr = inet_addr(host)) != INADDR_NONE) {
	hp = gethostbyaddr((const char *) &rem_sockaddr_in.sin_addr,
			   sizeof(rem_sockaddr_in.sin_addr), AF_INET);
	if (hp == NULL) {
	    /*
	     * herror("gethostbyaddr"); not an error for numeric
	     * addresses
	     */
	} else {
	    bcopy(hp->h_addr, (char *) &rem_sockaddr_in.sin_addr, hp->h_length);
	}
    } else {
	hp = gethostbyname(host);
	if (hp == NULL) {
	    herror("gethostbyname");
	    return (-1);
	}
	bcopy(hp->h_addr, (char *) &rem_sockaddr_in.sin_addr, hp->h_length);
    }
    sockt = socket(AF_INET, SOCK_STREAM, 0);
    if (sockt < 0) {
	perror("socket/tcp");
	return (-1);
    }
    port = htons(atoi(servp));
    if (!port) {
	sp = getservbyname(servp, "tcp");
	if (sp == 0) {
	    if (strcmp(servp, "rtcm-sc104") == 0) {
		sp = getservent();
		if (sp) {
		    sp->s_port = htons(2104);
		}
	    }
	}
	if (sp == 0) {
	    fprintf(stderr, "unknown service %s/%s\n", servp,
		    "tcp");
	    close(sockt);
	    return (-1);
	}
	port = sp->s_port;
    }
    rem_sockaddr_in.sin_len = sizeof(rem_sockaddr_in);
    rem_sockaddr_in.sin_family = AF_INET;
    rem_sockaddr_in.sin_port = port;

    if (connect(sockt, (const struct sockaddr *) & rem_sockaddr_in,
		sizeof(rem_sockaddr_in)) == -1) {
	if (verbose >= 5)
	    perror("connect");
	close(sockt);
	return (-1);
    }
    return (sockt);
}

/* end */
